@@ -14,6 +14,10 @@
 #include <sys/un.h>
 #endif
 
+#if defined(__NetBSD__)
+#include <sys/un.h>
+#endif
+
 #include "client.h"
 #include "linked_list.h"
 #include "log.h"
@@ -34,6 +38,23 @@ static int get_peer(int fd, pid_t *pid, uid_t *uid, gid_t *gid) {
 	*uid = cred.uid;
 	*gid = cred.gid;
 	return 0;
+#elif defined(__NetBSD__)
+	struct unpcbid cred;
+	socklen_t len = sizeof cred;
+	if (getsockopt(fd, 0, LOCAL_PEEREID, &cred, &len) == -1) {
+		// assume builtin backend
+		if (errno == EINVAL) {
+			*pid = getpid();
+			*uid = getuid();
+			*gid = getgid();
+			return 0;
+		}
+		return -1;
+	}
+	*pid = cred.unp_pid;
+	*uid = cred.unp_euid;
+	*gid = cred.unp_egid;
+	return 0;
 #elif defined(__FreeBSD__)
 	struct xucred cred;
 	socklen_t len = sizeof cred;
@@ -49,7 +70,7 @@ static int get_peer(int fd, pid_t *pid, uid_t *uid, gid_t *gid) {
 	*gid = cred.cr_ngroups > 0 ? cred.cr_groups[0] : (gid_t)-1;
 	return 0;
 #else
-	return -1;
+#error Unsupported platform
 #endif
 }
 
@@ -81,15 +102,22 @@ struct client *client_create(struct server *server, int client_fd) {
 
 void client_destroy(struct client *client) {
 	assert(client);
+
+#ifdef LIBSEAT
+	// The built-in backend version of seatd should terminate once its only
+	// client disconnects.
+	client->server->running = false;
+#endif
+
 	client->server = NULL;
 	if (client->connection.fd != -1) {
 		close(client->connection.fd);
 		client->connection.fd = -1;
 	}
-	linked_list_remove(&client->link);
 	if (client->seat != NULL) {
 		seat_remove_client(client);
 	}
+	linked_list_remove(&client->link);
 	if (client->event_source != NULL) {
 		event_source_fd_destroy(client->event_source);
 		client->event_source = NULL;
@@ -144,16 +172,12 @@ static int handle_open_seat(struct client *client) {
 		log_errorf("Could not find seat named %s", seat_name);
 		return -1;
 	}
-
 	if (seat_add_client(seat, client) == -1) {
 		log_errorf("Could not add client to target seat: %s", strerror(errno));
 		return -1;
 	}
-	linked_list_remove(&client->link);
-	linked_list_insert(&seat->clients, &client->link);
 
 	size_t seat_name_len = strlen(seat_name);
-
 	struct proto_server_seat_opened rmsg = {
 		.seat_name_len = (uint16_t)seat_name_len,
 	};
@@ -178,12 +202,7 @@ static int handle_close_seat(struct client *client) {
 		log_error("Protocol error: no seat associated with client");
 		return -1;
 	}
-
-	linked_list_remove(&client->link);
-	if (seat_remove_client(client) == -1) {
-		log_error("Could not remove client from seat");
-		return -1;
-	}
+	seat_remove_client(client);
 	linked_list_insert(&client->server->idle_clients, &client->link);
 
 	struct proto_header header = {
@@ -284,12 +303,22 @@ static int handle_switch_session(struct client *client, int session) {
 	}
 
 	if (seat_set_next_session(client, session) == -1) {
-		goto error;
+		goto fail;
+	}
+
+	struct proto_header header = {
+		.opcode = SERVER_SESSION_SWITCHED,
+		.size = 0,
+	};
+
+	if (connection_put(&client->connection, &header, sizeof header) == -1) {
+		log_errorf("Could not write response: %s", strerror(errno));
+		return -1;
 	}
 
 	return 0;
 
-error:
+fail:
 	return client_send_error(client, errno);
 }
 
@@ -300,12 +329,22 @@ static int handle_disable_seat(struct client *client) {
 	}
 
 	if (seat_ack_disable_client(client) == -1) {
-		goto error;
+		goto fail;
+	}
+
+	struct proto_header header = {
+		.opcode = SERVER_SEAT_DISABLED,
+		.size = 0,
+	};
+
+	if (connection_put(&client->connection, &header, sizeof header) == -1) {
+		log_errorf("Could not write response: %s", strerror(errno));
+		return -1;
 	}
 
 	return 0;
 
-error:
+fail:
 	return client_send_error(client, errno);
 }
 
@@ -461,7 +500,13 @@ int client_handle_connection(int fd, uint32_t mask, void *data) {
 			goto fail;
 		}
 		if (len == 0) {
+// https://man.netbsd.org/poll.2
+// Sockets produce POLLIN rather than POLLHUP when the remote end is closed.
+#if defined(__NetBSD__)
+			log_info("Client disconnected");
+#else
 			log_error("Could not read client connection: zero-length read");
+#endif
 			goto fail;
 		}
 
